'use strict';
/*
 This class implements interaction with UDF-compatible datafeed.
 See UDF protocol reference at
 https://github.com/tradingview/charting_library/wiki/UDF
 */

var Datafeeds = {};

Datafeeds.UDFCompatibleDatafeed = function (datafeedURL, socketClient, _enableLogging) {

    this._datafeedURL = datafeedURL;
    this._configuration = undefined;

    this._symbolSearch = null;
    this._symbolsStorage = null;
    this._barsPulseUpdater = null;

    if (socketClient) {
        this._barsPulseUpdater = new Datafeeds.DataPulseUpdater(this, socketClient);
        if (window.Settings.websocket_transport !== "centrifugo") {
            console.error('Сервис графика настроен только на centrifugo');
        }
    } else  {
        console.error('SocketClient must be passed into datafeed');
    }

    this._enableLogging = _enableLogging || false;
    this._initializationFinished = false;
    this._callbacks = {};

    this._initialize();
};

// Default Config
Datafeeds.UDFCompatibleDatafeed.prototype.defaultConfiguration = function () {
    return {
        supports_search: true,
        supports_group_request: false,
        supported_resolutions: ['1', '5', '15', '45', '30', '60', '120', '180', '240', 'D', 'W', 'M'],
        supports_marks: false,
        supports_timescale_marks: false,
        supports_time: true
    };
};

Datafeeds.UDFCompatibleDatafeed.prototype.getUpdater = function() {
    return this._barsPulseUpdater
};

Datafeeds.UDFCompatibleDatafeed.prototype.getServerTime = function (callback) {
    if (this._configuration.supports_time) {
        this._send(this._datafeedURL + '/time')
          .done(function (response) {
              callback(+response);
          })
          .fail(function () {

          });
    }
};

Datafeeds.UDFCompatibleDatafeed.prototype.on = function (event, callback) {

    if (!this._callbacks.hasOwnProperty(event)) {
        this._callbacks[event] = [];
    }

    this._callbacks[event].push(callback);
    return this;
};

Datafeeds.UDFCompatibleDatafeed.prototype._fireEvent = function (event, argument) {
    if (this._callbacks.hasOwnProperty(event)) {
        var callbacksChain = this._callbacks[event];
        for (var i = 0; i < callbacksChain.length; ++i) {
            callbacksChain[i](argument);
        }

        this._callbacks[event] = [];
    }
};

Datafeeds.UDFCompatibleDatafeed.prototype.onInitialized = function () {
    this._initializationFinished = true;
    this._fireEvent('initialized');
};

Datafeeds.UDFCompatibleDatafeed.prototype._logMessage = function (message) {
    if (this._enableLogging) {
        var now = new Date();
        console.log(now.toLocaleTimeString() + '.' + now.getMilliseconds() + '> ' + message);
    }
};

Datafeeds.UDFCompatibleDatafeed.prototype._send = function (url, data) {
    this._logMessage('New request: ' + url);
    return	$.get(url, data);
};

Datafeeds.UDFCompatibleDatafeed.prototype._initialize = function () {
    this._setupWithConfiguration(this.defaultConfiguration());
};

Datafeeds.UDFCompatibleDatafeed.prototype.onReady = function (callback) {
    var that = this;
    if (this._configuration) {
        setTimeout(function () {
            callback(that._configuration);
        }, 0);
    } else {
        this.on('configuration_ready', function () {
            callback(that._configuration);
        });
    }
};

Datafeeds.UDFCompatibleDatafeed.prototype._setupWithConfiguration = function (configurationData) {
    this._configuration = configurationData;

    if (!configurationData.exchanges) {
        configurationData.exchanges = [];
    }

    //	@obsolete; remove in 1.5
    var supported_resolutions = configurationData.supported_resolutions || configurationData.supported_resolutions;
    configurationData.supported_resolutions = supported_resolutions;

    //	@obsolete; remove in 1.5
    var symbolsTypes = configurationData.symbols_types || configurationData.symbolsTypes;
    configurationData.symbols_types = symbolsTypes;

    if (!configurationData.supports_search && !configurationData.supports_group_request) {
        throw 'Unsupported datafeed configuration. Must either support search, or support group request';
    }

    if (!configurationData.supports_search) {
        this._symbolSearch = new Datafeeds.SymbolSearchComponent(this);
    }

    if (configurationData.supports_group_request) {
        //	this component will call onInitialized() by itself
        this._symbolsStorage = new Datafeeds.SymbolsStorage(this);
    } else {
        this.onInitialized();
    }

    this._fireEvent('configuration_ready');
    this._logMessage('Initialized with ' + JSON.stringify(configurationData));
};

//	===============================================================================================================================
//	The functions set below is the implementation of JavaScript API.

Datafeeds.UDFCompatibleDatafeed.prototype.searchSymbols = function (searchString, exchange, type, onResultReadyCallback) {
    var MAX_SEARCH_RESULTS = 30;

    if (!this._configuration) {
        onResultReadyCallback([]);
        return;
    }

    if (this._configuration.supports_search) {

        this._send(this._datafeedURL + '/search', {
            //action: 'search',
            limit: MAX_SEARCH_RESULTS,
            query: searchString.toUpperCase(),
            type: type,
            exchange: exchange
        })
          .done(function (response) {

              for (var i = 0; i < response.length; ++i) {
                  if (!response[i].params) {
                      response[i].params = [];
                  }
              }

              if (typeof response.s == 'undefined' || response.s != 'error') {
                  onResultReadyCallback(response);
              } else {
                  onResultReadyCallback([]);
              }

          })
          .fail(function (reason) {
              onResultReadyCallback([]);
          });
    } else {

        if (!this._symbolSearch) {
            throw 'Datafeed error: inconsistent configuration (symbol search)';
        }

        var searchArgument = {
            searchString: searchString,
            exchange: exchange,
            type: type,
            onResultReadyCallback: onResultReadyCallback
        };

        if (this._initializationFinished) {
            this._symbolSearch.searchSymbols(searchArgument, MAX_SEARCH_RESULTS);
        } else {

            var that = this;

            this.on('initialized', function () {
                that._symbolSearch.searchSymbols(searchArgument, MAX_SEARCH_RESULTS);
            });
        }
    }
};

Datafeeds.UDFCompatibleDatafeed.prototype._symbolResolveURL = '/symbol';

Datafeeds.UDFCompatibleDatafeed.prototype.postProcessSymbolInfo = function (symbol) {
    return symbol;
};

//	BEWARE: this function does not consider symbol's exchange
Datafeeds.UDFCompatibleDatafeed.prototype.resolveSymbol = function (symbolName, onSymbolResolvedCallback, onResolveErrorCallback) {

    var that = this;

    if (!this._initializationFinished) {
        this.on('initialized', function () {
            that.resolveSymbol(symbolName, onSymbolResolvedCallback, onResolveErrorCallback);
        });

        return;
    }

    var resolveRequestStartTime = Date.now();
    that._logMessage('Resolve requested');

    function onResultReady(data) {
        var postProcessedData = data;
        if (that.postProcessSymbolInfo) {
            postProcessedData = that.postProcessSymbolInfo(postProcessedData);
        }

        that._logMessage('Symbol resolved: ' + (Date.now() - resolveRequestStartTime));

        onSymbolResolvedCallback(postProcessedData);
    }

    if (!this._configuration.supports_group_request) {
        this._send(this._datafeedURL + this._symbolResolveURL, {
            name: symbolName ? symbolName.replace('/', '_').toUpperCase() : ''
        })
          .done(function (response) {
              onResultReady(response);
              //TODO error processing
              //onResolveErrorCallback('unknown_symbol');
          })
          .fail(function (reason) {
              that._logMessage('Error resolving symbol: ' + JSON.stringify([reason]));
              onResolveErrorCallback('unknown_symbol');
          });
    } else {
        if (this._initializationFinished) {
            this._symbolsStorage.resolveSymbol(symbolName, onResultReady, onResolveErrorCallback);
        } else {
            this.on('initialized', function () {
                that._symbolsStorage.resolveSymbol(symbolName, onResultReady, onResolveErrorCallback);
            });
        }
    }
};

Datafeeds.UDFCompatibleDatafeed.prototype._historyURL = '/history';

Datafeeds.UDFCompatibleDatafeed.prototype.getBars = function (symbolInfo, resolution, rangeStartDate, rangeEndDate, onDataCallback, onErrorCallback) {

    //	timestamp sample: 1399939200
    if (rangeStartDate > 0 && (rangeStartDate + '').length > 10) {
        throw ['Got a JS time instead of Unix one.', rangeStartDate, rangeEndDate];
    }

    var that = this;

    var requestStartTime = Date.now();
    var symbol = symbolInfo.ticker.replace('/', '_').toUpperCase();

    this._send(this._datafeedURL + this._historyURL, {
        symbol: symbol,
        resolution: resolution,
        from: rangeStartDate,
        to: rangeEndDate
    })
      .done(function (response) {
          var nodata = response.s == 'no_data';

          if (!response.candles && !nodata) {
              if (!!onErrorCallback) {
                  onErrorCallback(response.s);
              }
              return;
          }

          var bars = [];

          //	data is JSON having format {s: "status" (ok, no_data, error),
          //  v: [volumes], t: [times], o: [opens], h: [highs], l: [lows], c:[closes], nb: "optional_unixtime_if_no_data"}
          var barsCount = nodata ? 0 : response.candles.length;

          var volumePresent = true; //typeof response.v != 'undefined';
          var ohlPresent = true; //typeof response.o != 'undefined';
          var candles = response.candles;

          for (var i = 0; i < barsCount; ++i) {

              var barValue = {
                  time:  candles[i].t,
                  close: candles[i].c
              };

              if (ohlPresent) {
                  barValue.open = candles[i].o;
                  barValue.high = candles[i].h;
                  barValue.low =  candles[i].l;
              } else {
                  barValue.open = barValue.high = barValue.low = barValue.close;
              }

              if (volumePresent) {
                  barValue.volume = candles[i].v;
              }


              bars.push(barValue);
          }

          onDataCallback(bars, {
              version: that._protocolVersion,
              noData: nodata,
              nextTime: response.nb || response.nextTime
          });
      })
      .fail(function (arg) {
          if (!!onErrorCallback) {
              onErrorCallback('network error: ' + JSON.stringify(arg));
          }
      });
};

Datafeeds.UDFCompatibleDatafeed.prototype.subscribeBars = function (symbolInfo, resolution, onRealtimeCallback, listenerGUID, onResetCacheNeededCallback) {
    this._barsPulseUpdater.subscribeDataListener(symbolInfo, resolution, onRealtimeCallback, listenerGUID);
    this._barsPulseUpdater.init(onRealtimeCallback, onResetCacheNeededCallback);
};

Datafeeds.UDFCompatibleDatafeed.prototype.unsubscribeBars = function (listenerGUID) {
    this._barsPulseUpdater.unsubscribeDataListener(listenerGUID);
    this._barsPulseUpdater.reset();
};

Datafeeds.UDFCompatibleDatafeed.prototype.calculateHistoryDepth = function (period, resolutionBack, intervalBack) {
    if (period === "1S") {
        return {
            // resolutionBack: 'D',
            intervalBack: 1000
        };
    }
    if (period === "30S") {
        return {
            intervalBack: 1000
        };
    }
};

//	==================================================================================================================================================
//	==================================================================================================================================================
//	==================================================================================================================================================

/*
 It's a symbol storage component for ExternalDatafeed. This component can
 * interact to UDF-compatible datafeed which supports whole group info requesting
 * do symbol resolving -- return symbol info by its name
 */
Datafeeds.SymbolsStorage = function (datafeed) {
    this._datafeed = datafeed;

    this._exchangesList = ['EXMO'];
    this._exchangesWaitingForData = {};
    this._exchangesDataCache = {};

    this._symbolsInfo = {};
    this._symbolsList = [];

    this._requestFullSymbolsList();
};

Datafeeds.SymbolsStorage.prototype._requestFullSymbolsList = function () {

    var that = this;
    var datafeed = this._datafeed;

    for (var i = 0; i < this._exchangesList.length; ++i) {

        var exchange = this._exchangesList[i];

        if (this._exchangesDataCache.hasOwnProperty(exchange)) {
            continue;
        }

        this._exchangesDataCache[exchange] = true;

        this._exchangesWaitingForData[exchange] = 'waiting_for_data';

        this._datafeed._send(this._datafeed._datafeedURL + '/symbolInfo', {
            group: exchange
        })
          .done(function (exchange) {
              return function (response) {
                  that._onExchangeDataReceived(exchange, JSON.parse(response));
                  that._onAnyExchangeResponseReceived(exchange);
              };
          }(exchange)) //jshint ignore:line
          .fail(function (exchange) {
              return function (reason) {
                  that._onAnyExchangeResponseReceived(exchange);
              };
          }(exchange)); //jshint ignore:line
    }
};

Datafeeds.SymbolsStorage.prototype._onExchangeDataReceived = function (exchangeName, data) {

    function tableField(data, name, index) {
        return data[name] instanceof Array ?
          data[name][index] :
          data[name];
    }

    try {
        for (var symbolIndex = 0; symbolIndex < data.symbol.length; ++symbolIndex) {

            var symbolName = data.symbol[symbolIndex];
            var listedExchange = tableField(data, 'exchange_listed', symbolIndex);
            var tradedExchange = tableField(data, 'exchange_traded', symbolIndex);
            var fullName = tradedExchange + ':' + symbolName;

            //	This feature support is not implemented yet
            //	var hasDWM = tableField(data, "has-dwm", symbolIndex);

            var hasIntraday = tableField(data, 'has-intraday', symbolIndex);

            var tickerPresent = typeof data.ticker != 'undefined';

            var symbolInfo = {
                name: symbolName,
                base_name: [listedExchange + ':' + symbolName],
                description: tableField(data, 'description', symbolIndex),
                full_name: fullName,
                legs: [fullName],
                has_intraday: hasIntraday,
                has_no_volume: tableField(data, 'has-no-volume', symbolIndex),
                listed_exchange: listedExchange,
                exchange: tradedExchange,
                minmov: tableField(data, 'minmovement', symbolIndex) || tableField(data, 'minmov', symbolIndex),
                minmove2: tableField(data, 'minmove2', symbolIndex) || tableField(data, 'minmov2', symbolIndex),
                fractional: tableField(data, 'fractional', symbolIndex),
                pointvalue: tableField(data, 'pointvalue', symbolIndex),
                pricescale: tableField(data, 'pricescale', symbolIndex),
                type: tableField(data, 'type', symbolIndex),
                session: tableField(data, 'session-regular', symbolIndex),
                ticker: tickerPresent ? tableField(data, 'ticker', symbolIndex) : symbolName,
                timezone: tableField(data, 'timezone', symbolIndex),
                supported_resolutions: tableField(data, 'supported-resolutions', symbolIndex) || this._datafeed.defaultConfiguration().supported_resolutions,
                force_session_rebuild: tableField(data, 'force-session-rebuild', symbolIndex) || false,
                has_daily: tableField(data, 'has-daily', symbolIndex) || true,
                intraday_multipliers: tableField(data, 'intraday-multipliers', symbolIndex) || ['1', '5', '15', '30', '60'],
                has_fractional_volume: tableField(data, 'has-fractional-volume', symbolIndex) || false,
                has_weekly_and_monthly: tableField(data, 'has-weekly-and-monthly', symbolIndex) || false,
                has_empty_bars: tableField(data, 'has-empty-bars', symbolIndex) || false,
                volume_precision: tableField(data, 'volume-precision', symbolIndex) || 0
            };

            this._symbolsInfo[symbolInfo.ticker] = this._symbolsInfo[symbolName] = this._symbolsInfo[fullName] = symbolInfo;
            this._symbolsList.push(symbolName);
        }
    }
    catch (error) {
        throw 'API error when processing exchange `' + exchangeName + '` symbol #' + symbolIndex + ': ' + error;
    }
};

Datafeeds.SymbolsStorage.prototype._onAnyExchangeResponseReceived = function (exchangeName) {

    delete this._exchangesWaitingForData[exchangeName];

    var allDataReady = Object.keys(this._exchangesWaitingForData).length === 0;

    if (allDataReady) {
        this._symbolsList.sort();
        this._datafeed._logMessage('All exchanges data ready');
        this._datafeed.onInitialized();
    }
};

//	BEWARE: this function does not consider symbol's exchange
Datafeeds.SymbolsStorage.prototype.resolveSymbol = function (symbolName, onSymbolResolvedCallback, onResolveErrorCallback) {
    var that = this;

    setTimeout(function () {
        if (!that._symbolsInfo.hasOwnProperty(symbolName)) {
            onResolveErrorCallback('invalid symbol');
        } else {
            onSymbolResolvedCallback(that._symbolsInfo[symbolName]);
        }
    }, 0);
};

//	==================================================================================================================================================
//	==================================================================================================================================================
//	==================================================================================================================================================

/*
 It's a symbol search component for ExternalDatafeed. This component can do symbol search only.
 This component strongly depends on SymbolsDataStorage and cannot work without it. Maybe, it would be
 better to merge it to SymbolsDataStorage.
 */

Datafeeds.SymbolSearchComponent = function (datafeed) {
    this._datafeed = datafeed;
};

//	searchArgument = { searchString, onResultReadyCallback}
Datafeeds.SymbolSearchComponent.prototype.searchSymbols = function (searchArgument, maxSearchResults) {

    if (!this._datafeed._symbolsStorage) {
        throw 'Cannot use local symbol search when no groups information is available';
    }

    var symbolsStorage = this._datafeed._symbolsStorage;

    var results = []; // array of WeightedItem { item, weight }
    var queryIsEmpty = !searchArgument.searchString || searchArgument.searchString.length === 0;
    var searchStringUpperCase = searchArgument.searchString.toUpperCase();

    for (var i = 0; i < symbolsStorage._symbolsList.length; ++i) {
        var symbolName = symbolsStorage._symbolsList[i];
        var item = symbolsStorage._symbolsInfo[symbolName];

        if (searchArgument.type && searchArgument.type.length > 0 && item.type != searchArgument.type) {
            continue;
        }

        if (searchArgument.exchange && searchArgument.exchange.length > 0 && item.exchange != searchArgument.exchange) {
            continue;
        }

        var positionInName = item.name.toUpperCase().indexOf(searchStringUpperCase);
        var positionInDescription = item.description.toUpperCase().indexOf(searchStringUpperCase);

        if (queryIsEmpty || positionInName >= 0 || positionInDescription >= 0) {
            var found = false;
            for (var resultIndex = 0; resultIndex < results.length; resultIndex++) {
                if (results[resultIndex].item == item) {
                    found = true;
                    break;
                }
            }

            if (!found) {
                var weight = positionInName >= 0 ? positionInName : 8000 + positionInDescription;
                results.push({item: item, weight: weight});
            }
        }
    }

    searchArgument.onResultReadyCallback(
      results
        .sort(function (weightedItem1, weightedItem2) {
            return weightedItem1.weight - weightedItem2.weight;
        })
        .map(function (weightedItem) {
            var item = weightedItem.item;
            return {
                symbol: item.name,
                full_name: item.full_name,
                description: item.description,
                exchange: item.exchange,
                params: [],
                type: item.type,
                ticker: item.name
            };
        })
        .slice(0, Math.min(results.length, maxSearchResults))
    );
};

//	==================================================================================================================================================
//	==================================================================================================================================================
//	==================================================================================================================================================

/*
 */

Datafeeds.DataPulseUpdater = function (datafeed, socketClient) {
    this._datafeed = datafeed;
    this._subscribers = {};
    this._socketClient = socketClient;
    this._updatedResolution = ["1", "5", "15", "30", "45", "60"];
    this._meesageQueue = [];
    this._updateFn = null;
    var self = this;
    this.intervalId = 0;

    if (window.Settings.websocket_transport == "centrifugo") {
        // удалить  https://github.com/centrifugal/centrifuge-js/issues/46
        this._socketClient.on('connect', function (context) {
            if (Object.keys(self._subscribers).length) {
                Object.entries(self._subscribers).forEach(function (sub) {
                    self._socketClient._subscribe(sub[1].listener)
                })
            }
        });
    }
};

Datafeeds.DataPulseUpdater.prototype.reset = function () {
    this._meesageQueue = [];
    clearInterval(this.intervalId);
};

Datafeeds.DataPulseUpdater.prototype.init = function (_updateFn, _onResetCacheNeededCallback) {
    clearInterval(this.intervalId);
    this._meesageQueue = [];
    this._updateFn = _updateFn;
    this.resetCache = _onResetCacheNeededCallback;
    var self = this;

    this.intervalId = setInterval(function () {
        if (self._updateFn) {
            for (var i = 0; i < self._meesageQueue.length; i++) {
                var item = self._meesageQueue[i];
                self._updateFn(item);
            }
            self._meesageQueue.length = 0;
        }
    }, 200)
};


Datafeeds.DataPulseUpdater.prototype.unsubscribeDataListener = function (listenerGUID) {
    this._datafeed._logMessage('Unsubscribing ' + listenerGUID);

    if (this._subscribers[listenerGUID]) {
        this.unsub(this._subscribers[listenerGUID].listener);
        delete this._subscribers[listenerGUID];
    }
};

Datafeeds.DataPulseUpdater.prototype.subscribeDataListener = function (symbolInfo, resolution, newDataCallback, listenerGUID) {

    this._datafeed._logMessage('Subscribing ' + listenerGUID);
    var prefix = '';

    //if (resolution !== '30') prefix = "_" + resolution;
    prefix = "_" + resolution;

    if (!~this._updatedResolution.indexOf(resolution))
        return;

    var key = symbolInfo.name.replace('/','_') + prefix;

    if (!this._subscribers.hasOwnProperty(listenerGUID)) {

        this._subscribers[listenerGUID] = {
            subscribtionId: key,
            symbolInfo: symbolInfo,
            resolution: resolution,
            lastBarTime: NaN,
            listener: []
        };
    }

    this._subscribers[listenerGUID].listener = this.makeSubscription(key, newDataCallback);
};

Datafeeds.DataPulseUpdater.prototype.periodLengthSeconds = function (resolution, requiredPeriodsCount) {
    var daysCount = 0;

    if (resolution == 'D') {
        daysCount = requiredPeriodsCount;
    } else if (resolution == 'M') {
        daysCount = 31 * requiredPeriodsCount;
    } else if (resolution == 'W') {
        daysCount = 7 * requiredPeriodsCount;
    } else {
        daysCount = requiredPeriodsCount * resolution / (24 * 60);
    }

    return daysCount * 24 * 60 * 60;
};

Datafeeds.DataPulseUpdater.prototype.makeSubscription = function (channel, callback) {
    var sub = null;
    var noop = function () {};

    if (window.Settings.websocket_transport == "pusher") {
        sub = this._socketClient.subscribe(channel);
        sub.bind("update", function (message) {
            callback(message.data)
        });

        return {
            subscription: sub,
            channel: channel
        };
    }

    if (window.Settings.websocket_transport == "centrifugo") {
        var self = this;
        return this._socketClient.subscribe(channel, function (message) {
            self._meesageQueue.push(message.data);
        })
    }

    throw new Error('Unsupported websocket transport')
};

Datafeeds.DataPulseUpdater.prototype.unsub = function (unsabObject) {
    var transport = window.Settings.websocket_transport;

    switch (transport) {
        case "pusher": this._socketClient.unsubscribe(unsabObject.channel); break;
        case "centrifugo": unsabObject.unsubscribe(); break;
        default: throw new Error('Unsupported websocket transport')
    }
}